import os
import yaml
from box import Box

class KACLConfig:
    def __init__(self, config_file=None):
        default_config_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config', 'kacl-default.yml')
        with open(default_config_file, 'r') as df:
            self.__config = yaml.safe_load(df)
        if config_file:
            with open(os.path.expanduser(config_file), 'r') as f:
                KACLConfig.merge(self.__config, yaml.safe_load(f))

        self.__config = Box(self.__config)

    def allowed_header_titles(self):
        return self.__config.kacl.allowed_header_titles

    def allowed_version_sections(self):
        return self.__config.kacl.allowed_version_sections

    def default_content(self):
        return self.__config.kacl.default_content

    def git_tag_name(self):
        return self.__config.kacl.git.tag_name

    def git_tag_description(self):
        return self.__config.kacl.git.tag_description

    def changelog_file_path(self):
        return self.__config.kacl.file

    def set_changelog_file_path(self, file_path):
        self.__config.kacl.file = file_path

    def git_create_tag(self):
        return self.__config.kacl.git.tag

    def git_create_commit(self):
        return self.__config.kacl.git.commit

    def git_commit_message(self):
        return self.__config.kacl.git.commit_message

    def git_commit_additional_files(self):
        return self.__config.kacl.git.commit_additional_files

    def link_auto_generate(self):
        if 'auto_generate' in self.__config.kacl.links:
            return self.__config.kacl.links.auto_generate

    def link_host_url(self):
        if 'host_url' in self.__config.kacl.links:
            return self.__config.kacl.links.host_url

    def links_compare_versions_template(self):
        if 'compare_versions_template' in self.__config.kacl.links:
            return self.__config.kacl.links.compare_versions_template

    def links_unreleased_changes_template(self):
        if 'unreleased_changes_template' in self.__config.kacl.links:
            return self.__config.kacl.links.unreleased_changes_template

    def links_initial_version_template(self):
        if 'initial_version_template' in self.__config.kacl.links:
            return self.__config.kacl.links.initial_version_template


    @staticmethod
    def merge(a, b, path=None):
        """ merge two dictionaries
        :param a: dictionary the values will be merged into
        :param b: dictionary the values will be used when overwriting values
        :param path:
        :return:
        """
        if path is None:
            path = []
        for key in b:
            if key in a:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    KACLConfig.merge(a[key], b[key], path + [str(key)])
                elif a[key] == b[key]:
                    pass  # same leaf value
                else:
                    a[key] = b[key]

            else:
                a[key] = b[key]
        return a


from .element import KACLElement

import re

class KACLParser:
    semver_regex = r'(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?'

    @staticmethod
    def parse_header(text, start_depth, end_depth=None, line_offset=0):
        if not end_depth:
            end_depth = start_depth

        elements = []
        reg_expr_start = r'(\n{depth}|^{depth})\s+?(.*)\n'.format(depth="#"*start_depth)
        reg_expr_end = r'(\n{depth}|^{depth})\s+?(.*)\n'.format(depth="#"*end_depth)
        for match in re.finditer(reg_expr_start, text):
            # find end of section
            raw = match.group().strip()
            title = match.group(2).strip()
            start = match.start()
            end = match.end()
            line_number = text[:start].count('\n')+line_offset
            if match.group()[0] == '\n':
                line_number += 2
            else:
                line_number += 1

            next_match = re.search(reg_expr_end, text[end:])
            body = None
            if next_match:
                body = text[end:next_match.start()+end]
            else:
                body = text[end:]
            elements.append(KACLElement(raw=raw, title=title,
                                        body=body, line_number=line_number))

        return elements

    @staticmethod
    def parse_link_references(text):
        link_references = dict()
        begin = None
        reg_expr = r'\n\[(.*)\]:(.*)'
        for match in re.finditer(reg_expr, text):
            if begin is None:
                begin = match.start()
            version = match.group(1).strip()
            link = match.group(2).strip()
            line_number = text[:match.start()].count('\n')+1
            link_references[version] = KACLElement(raw=match.group().strip(), title=version, body=link, line_number=line_number)

        return begin, link_references

    @staticmethod
    def parse_sem_ver(text, regex=None):
        if regex == None:
            regex = KACLParser.semver_regex
        m = re.search(regex, text)
        if m:
            return m.group().strip()



# -*- coding: utf-8 -*-

"""kacl.kacl-cli: provides entry point main()."""

import sys
import os
import kacl
import chalk
import click
import json
import semver
import traceback
import git
from datetime import datetime

from kacl.exception import KACLException

def load_changelog(ctx):
    config = ctx.obj['config']
    file = ctx.obj['file']

    changelog_file = os.path.join(os.getcwd(), file)
    if not os.path.exists(changelog_file):
        click.echo(click.style("Error: ", fg='red') +
                    f"{changelog_file} not found")
        sys.exit(1)

    if config is None:
        default_config_path = os.path.join(os.getcwd(), '.kacl.yml')
        config = default_config_path if os.path.exists(default_config_path) else None

    kacl_config = kacl.KACLConfig(config)
    if file:
        kacl_config.set_changelog_file_path(file)

    # read the changelog
    kacl_changelog = kacl.load(kacl_config.changelog_file_path())
    kacl_changelog.set_config(kacl_config)

    # share the objects
    return (kacl_changelog, changelog_file)

def prefixed_environ():
    return dict((("${}".format(key), value) for key, value in os.environ.items()))

@click.group(invoke_without_command=True)
@click.option('-v', '--version', is_flag=True, required=False, help='Prints the current version of the CLI.')
@click.option('-c', '--config', required=False, default=None, type=click.Path(exists=False), help='Path to kacl config file.', show_default=True)
@click.option('-f', '--file', required=False, default='CHANGELOG.md', type=click.Path(exists=False), help='Path to changelog file.', show_default=True)
@click.pass_context
def cli(ctx, version=None, config=None, file=None):
    if ctx.obj is None:
        ctx.obj = dict()

    ctx.obj['config'] = config
    ctx.obj['file'] = file

    # if --version was given, print version and exit directly
    if version:
        click.echo(kacl.__version__)
        sys.exit(0)


@cli.command()
@click.pass_context
@click.argument('section', type=str)
@click.argument('message', type=str)
@click.option('-m', '--modify', is_flag=True, help='This option will add the changes directly into changelog file.')
def add(ctx, section, message, modify):
    """Adds a given message to a specified unreleased section. Use '--modify' to directly modify the changelog file.
    """
    kacl_changelog, kacl_changelog_filepath = load_changelog(ctx)

    # add changes to changelog
    kacl_changelog.add(section=section, data=message)
    kacl_changelog_content = kacl.dump(kacl_changelog)
    if modify:
        with open(kacl_changelog_filepath, 'w') as f:
            f.write(kacl_changelog_content)
        f.close()
    else:
        click.echo(kacl_changelog_content)


@cli.command()
@click.pass_context
def current(ctx):
    """Returns the current version from the Changelog.
    """
    kacl_changelog, kacl_changelog_filepath = load_changelog(ctx)

    current_version = kacl_changelog.current_version()
    click.echo(current_version)


@cli.command()
@click.pass_context
@click.argument('version', type=str)
def get(ctx, version):
    """Returns a given version from the Changelog.
    """
    kacl_changelog, kacl_changelog_filepath = load_changelog(ctx)

    # add changes to changelog
    kacl_version = kacl_changelog.get(version)

    if kacl_version:
        kacl_changelog_content = kacl.dump(kacl_version)
        click.echo(kacl_changelog_content)
    else:
        click.echo(click.style("Error: ", fg='red') +
                   f'"{version}" could not be found in changelog.')
        sys.exit(1)


@cli.group(invoke_without_command=True)
@click.pass_context
def link(ctx):
    """Lets you modify version links
    """
    pass

@link.command()
@click.pass_context
@click.option('-m', '--modify', is_flag=True, help='This option will add the changes directly into changelog file.')
@click.option('--host-url', required=False, default=None, type=str, help='Host url to the git service. (i.e https://github.com/mschmieder/python-kacl)', show_default=True)
@click.option('--compare-versions-template', required=False, default=None, type=str, help='Template string for version comparison link.', show_default=True)
@click.option('--unreleased-changes-template', required=False, default=None, type=str, help='Template string for unreleased changes link.', show_default=True)
@click.option('--initial-version-template', required=False, default=None, type=str, help='Template string for initial version link.', show_default=True)
def generate(ctx, modify, host_url, compare_versions_template, unreleased_changes_template, initial_version_template):
    kacl_changelog, kacl_changelog_filepath = load_changelog(ctx)

    kacl_changelog.generate_links(host_url=host_url,
                                  compare_versions_template=compare_versions_template,
                                  unreleased_changes_template=unreleased_changes_template,
                                  initial_version_template=initial_version_template)

    kacl_changelog_content = kacl.dump(kacl_changelog)
    if modify:
        with open(kacl_changelog_filepath, 'w') as f:
            f.write(kacl_changelog_content)
        f.close()
    else:
        click.echo(kacl_changelog_content)


@cli.command()
@click.pass_context
@click.option('--json', 'as_json', is_flag=True, help='Print validation output as json.')
def verify(ctx, as_json):
    """Veryfies if the changelog is in "keep-a-changelog" format.
    Use '--json' get JSON formatted output that can be easier integrated into CI workflows.
    Exit code is the number of identified errors.
    """
    kacl_changelog, kacl_changelog_filepath = load_changelog(ctx)
    kacl_changelog_filepath = os.path.basename(kacl_changelog_filepath)

    valid = kacl_changelog.is_valid()
    validation = kacl_changelog.validate()
    if as_json:
        validation_map = validation.convert_to_dict()
        click.echo(json.dumps(validation_map, sort_keys=True, indent=4))
    else:
        green = chalk.Chalk('green')
        red = chalk.Chalk('red')
        white = chalk.Chalk('white')
        bold = chalk.bold

        for error in validation.errors():
            start_char_pos, end_char_pos = error.position()

            char_indicator = start_char_pos
            if start_char_pos == None:
                char_indicator = 0

            click.echo(bold + kacl_changelog_filepath + ':' +
                       f'{error.line_number()}:{char_indicator}: ' +
                       red + 'error: ' +
                       white + error.error_message() +
                       chalk.RESET)

            if error.line():
                click.echo(error.line())
                if start_char_pos != None and end_char_pos != None:
                    mark_length = end_char_pos-start_char_pos-1
                    click.echo(' '*start_char_pos +
                               green + '^' +
                               '~'*(mark_length) +
                               chalk.RESET)
        if not valid:
            click.echo(f'{len(validation.errors())} error(s) generated.')
        else:
            click.secho('Success', fg='green')

    if not valid:
        sys.exit(len(validation.errors()))


@cli.command()
@click.pass_context
@click.argument('version', type=str)
@click.option('-m', '--modify', is_flag=True, help='This option will add the changes directly into changelog file.')
@click.option('-l', '--link', required=False, default=None, type=str, help='A url that the version will be linked with.', show_default=True)
@click.option('-g', '--auto-link', is_flag=True, help='Will automatically create and update necessary links.')
@click.option('-c', '--commit', is_flag=True, help='If passed this will create a git commit with the changed Changelog.')
@click.option('--no-commit', is_flag=True, help='This will disabled the commit option even if it is set in the config.')
@click.option('--commit-message', required=False, default=None, type=str, help='The commit message to use when using --commit flag')
@click.option('-t', '--tag', is_flag=True, help='If passed this will create a git tag for the newly released version.')
@click.option('--tag-name', required=False, default=None, type=str, help='The tag name to use when using --tag flag')
@click.option('--tag-description', required=False, default=None, type=str, help='The tag description text to use when using --tag flag')
@click.option('-d', '--allow-dirty', is_flag=True, help='If passed this will allow to commit/tag even on a "dirty".')
def release(ctx, version, modify, link, auto_link, commit, no_commit, commit_message, tag, tag_name, tag_description, allow_dirty):
    """Creates a release for the latest 'unreleased' changes. Use '--modify' to directly modify the changelog file.
    You can automatically use the latest version by using the version keywords 'major', 'minor', 'patch'

    Example:

        kacl-cli release 1.0.0

        kacl-cli release major|minor|patch
    """
    kacl_changelog, kacl_changelog_filepath = load_changelog(ctx)
    kacl_config = kacl_changelog.config()

    if not kacl_changelog.is_valid():
        click.echo(click.style("Error: ", fg='red') +
                       f"Changelog is not valid. Run 'kacl-cli verify' for more information.")
        sys.exit(1)

    # check if the version string indicates automatic increment
    increment = None
    if version in ['major', 'minor', 'patch']:
        increment = version
        version = None  # reset
    else:
        # check if version is a valid semantic version
        try:
            semver.parse(version)
        except:
            click.echo(click.style("Error: ", fg='red') +
                       f'"{version}" not a valid semantic version.')
            sys.exit(1)

    if not kacl_changelog.has_changes():
        click.echo(click.style("Error: ", fg='red') +
                   'The current changelog has no changes. You can only release if changes are available.')
        sys.exit(1)

    # get the latest_version before the release
    latest_version = kacl_changelog.current_version()

    # check config to see if we need to autogenerate changes
    if auto_link is False and kacl_config.link_auto_generate():
        auto_link = True

    kacl_changelog.release(version=version, link=link, auto_link=auto_link, increment=increment)

    # get the new version
    new_version = kacl_changelog.current_version()

    # dump the content
    kacl_changelog_content = kacl.dump(kacl_changelog)

    # check if we should modify the file
    if modify:
        with open(kacl_changelog_filepath, 'w') as f:
            f.write(kacl_changelog_content)
        f.close()

        if not no_commit:
            if commit or tag or kacl_config.git_create_commit() or kacl_config.git_create_tag():
                vcs_context = {
                    "latest_version": latest_version,
                    "new_version": new_version,
                }
                time_context = {
                    'now': datetime.now(),
                    'utcnow': datetime.utcnow(),
                }
                vcs_context.update(time_context)
                vcs_context.update(prefixed_environ())

                commit_message  = commit_message if commit_message != None else kacl_config.git_commit_message()
                tag_name        = tag_name if tag_name != None else kacl_config.git_tag_name()
                tag_description = tag_description if tag_description != None else kacl_config.git_tag_description()

                repo = None
                try:
                    repo = git.Repo(os.getcwd())
                except git.InvalidGitRepositoryError:
                    click.echo(click.style("Error: ", fg='red') +
                            f'"{os.getcwd()}" is no valid git repository.')

                if repo:
                    if not repo.is_dirty() and not allow_dirty:
                        click.echo(click.style("Error: ", fg='red') +
                                f"Repository is marked 'dirty'. Use --allow-dirty if you want to commit/tag on a dirty repository")

                    if commit or kacl_config.git_create_commit():
                        repo.git.add(kacl_config.changelog_file_path())
                        for f in kacl_config.git_commit_additional_files():
                            repo.git.add(f)
                        repo.git.commit('-m', commit_message.format(**vcs_context))

                    if tag or kacl_config.git_create_tag():
                        repo.create_tag(tag_name.format(**vcs_context),
                                        message=tag_description.format(**vcs_context))
    else:
        click.echo(kacl_changelog_content)


@cli.command()
@click.option('-o', '--output-file', required=False, type=click.Path(exists=False), help='File to write the created changelog to.')
def new(output_file):
    """Creates a new changelog.
    """
    kacl_changelog = kacl.new()
    kacl_changelog_content = kacl.dump(kacl_changelog)
    if output_file:
        with open(output_file, 'w') as f:
            f.write(kacl_changelog_content)
        f.close()
    else:
        click.echo(kacl_changelog_content)


def start():
    try:
        cli(obj={})
    except SystemExit as e:
        sys.exit(e.code)
    except KACLException as e:
        click.secho(str(e), fg='red')
        sys.exit(1)
    except:
        click.echo(traceback.format_exc())
        sys.exit(1)

if __name__ == '__main__':
    start()



from .document import KACLDocument
from .element import KACLElement
from .version import KACLVersion
from .changes import KACLChanges

class KACLMarkdownSerializer:
    def __init__(self):
        pass

    def serialize(self, document):
        if isinstance(document, KACLDocument):
            data = [ self.__serialize_header(document.header()),
                    document.header().body() ]

            link_references = []
            for version in document.versions():
                data.extend([ self.__serialize_version(version), '' ])
                if version.has_link_reference():
                    link_references.append(self.__serialize_link_reference(version))

            data.extend(link_references)
            return '\n'.join(data)
        elif isinstance(document, KACLVersion):
            return self.__serialize_version(document)

    def __serialize_header(self, obj):
        if isinstance(obj, KACLChanges):
            return f'### {obj.title()}'
        elif isinstance(obj, KACLVersion):
            version_decorator_left = ''
            version_decorator_right = ''
            if obj.has_link_reference():
                version_decorator_left = '['
                version_decorator_right = ']'
            if obj.date():
                return f'## {version_decorator_left}{obj.version()}{version_decorator_right} - {obj.date()}'
            else:
                return f'## {version_decorator_left}{obj.version()}{version_decorator_right}'
        elif isinstance(obj, KACLElement):
            return f'# {obj.title()}'

    def __serialize_version(self, obj):
        lines = [ self.__serialize_header(obj) ]
        for title, changes in obj.sections().items():
            lines.extend([
                self.__serialize_header(changes),
                self.__serialize_list(changes.items()),
                ''
            ])

        return '\n'.join(lines).strip()

    def __serialize_list(self, obj):
        lines = [ f'- {x}' for x in obj ]
        return '\n'.join(lines)

    def __serialize_link_reference(self, obj):
        return f'[{obj.version()}]: {obj.link()}'


class KACLValidationError():
    def __init__(self, line="", line_number=0, start_character_pos=None, end_character_pos=None, error_message=""):
        self.__line_number = line_number
        self.__start_character_pos = start_character_pos
        self.__end_character_pos = end_character_pos
        self.__error_message = error_message
        self.__line = line

    def line_number(self):
        return self.__line_number

    def position(self):
        return self.__start_character_pos, self.__end_character_pos

    def line(self):
        return self.__line

    def error_message(self):
        return self.__error_message


class KACLValidation():
    def __init__(self):
        self.__validation_errors = []

    def is_valid(self):
        return (len(self.__validation_errors) == 0)

    def errors(self):
        return self.__validation_errors

    def add_error(self, line, line_number, error_message, start_character_pos=None, end_character_pos=None):
        self.__validation_errors.append(KACLValidationError(line=line,
                                                            line_number=line_number,
                                                            start_character_pos=start_character_pos,
                                                            end_character_pos=end_character_pos,
                                                            error_message=error_message))


    def convert_to_dict(self):
        validation_map = dict()
        validation_map['valid'] = (len(self.__validation_errors) == 0)
        errors = []
        for error in self.__validation_errors:
            error_map = {
                "line": error.line(),
                "line_number": error.line_number(),
                "start_char_pos": error.position()[0],
                "end_character_pos": error.position()[1],
                "error_message": error.error_message()
            }
            errors.append(error_map)
        validation_map['errors'] = errors
        return validation_map



from .element import KACLElement
from .changes import KACLChanges
from .parser import KACLParser

import re
import semver


class KACLVersion(KACLElement):
    def __init__(self, element=KACLElement(), version="", date="", sections=None, link=None):
        KACLElement.__init__(self,
                             raw=element.raw(),
                             title=element.title(),
                             body=element.body(),
                             line_number=element.line_number())
        self.__date = date
        self.__version = version
        if sections is None:
            self.__sections = dict()
        else:
            self.__sections = sections
        self.__link_reference = None
        self.set_link(link)

    def link(self):
        if self.__link_reference:
            return self.__link_reference.body()

    def set_link(self, link):
        if isinstance(link, KACLElement):
            self.__link_reference = link
        elif link != None:
            self.__link_reference = KACLElement(title=self.version(), body=link)

    def has_link_reference(self):
        if not self.__link_reference:
            return False
        else:
            return (self.__link_reference.body() != None and len(self.__link_reference.body()))

    def date(self):
        if not len(self.__date):
            title = self.title()
            m = re.search(r'\d\d\d\d-\d\d-\d\d', title)
            if m:
                self.__date = m.group().strip()

        return self.__date

    def version(self):
        if not len(self.__version):
            title = self.title()
            version = KACLParser.parse_sem_ver(title)
            if version:
                self.__version = version
            elif 'unreleased' in title.lower():
                self.__version = "Unreleased"

        return self.__version

    def semver(self):
        return semver.parse(self.version())

    def set_version(self, version):
        self.__version = version

    def sections(self):
        if not len(self.__sections) and len(self.body().strip()):
            self.__sections = dict()
            sections = KACLParser.parse_header(text=self.body(),
                                               start_depth=3,
                                               end_depth=3,
                                               line_offset=self.line_number())
            for section in sections:
                sec = KACLChanges(section)
                self.__sections[sec.title()] = sec
        return self.__sections

    def changes(self, section):
        sections = self.sections()
        if sections and section in sections:
            return sections[section]

        return None

    def add(self, section, change):
        if section not in self.sections():
            self.__sections[section] = KACLChanges(KACLElement(
                title=section, body="", line_number=None))
        self.__sections[section].add(change)



from .parser import KACLParser
from .element import KACLElement
import re


class KACLChanges(KACLElement):
    def __init__(self, element):
        KACLElement.__init__(self, 
                             raw=element.raw(),
                             title=element.title(),
                             body=element.body(),
                             line_number=element.line_number())
        self.__items = []

    def items(self):
        if not len(self.__items) and len(self.body().strip()):
            body = self.body()
            items = ("\n"+body).split('\n-')
            self.__items = [x.strip() for x in items if len(x.strip()) > 0]
        return self.__items

    def add(self, item):
        self.items().append(item)



import re

class LinkProvider:
    def __init__(self, host_url=None, compare_versions_template=None, unreleased_changes_template=None, initial_version_template=None):
        self.host_url = self.__sanatize_url(host_url)
        self.compare_versions_template = compare_versions_template
        self.unreleased_changes_template = unreleased_changes_template
        self.initial_version_template = initial_version_template

    def __sanatize_url(self, url):
        if url:
            regExpr = r'git@(.*):(.*).git'
            m = re.search(regExpr, url)
            if m:
                return f'https://{m.group(1)}/{m.group(2)}'
        return url

    def compare_versions(self, version=None, previous_version=None, latest_version=None):
        return self.compare_versions_template.format(
            host=self.host_url,
            version=version,
            previous_version=previous_version,
            latest_version=latest_version,
        )

    def initial_version(self, version=None, previous_version=None, latest_version=None):
        return self.initial_version_template.format(
            host=self.host_url,
            version=version,
            previous_version=previous_version,
            latest_version=latest_version,
        )

    def unreleased_changes(self, version=None, previous_version=None, latest_version=None):
        return self.unreleased_changes_template.format(
            host=self.host_url,
            version=version,
            previous_version=previous_version,
            latest_version=latest_version,
        )


class KACLElement:
    def __init__(self, raw="", title="", body="", line_number=None):
        self.__raw = raw
        self.__title = title
        self.__body = body
        self.__line_number = line_number

    def line_number(self):
        return self.__line_number

    def title(self):
        return self.__title

    def body(self):
        return self.__body

    def raw(self):
        return self.__raw


import datetime
import re
import semver
import os
import git

from .element import KACLElement
from .version import KACLVersion
from .parser import KACLParser
from .config import KACLConfig
from .link_provider import LinkProvider
from .validation import KACLValidation
from .exception import KACLException

WINDOWS_LINE_ENDING = r'\r\n'
UNIX_LINE_ENDING = r'\n'

class KACLDocument:
    def __init__(self, data="", headers=[], versions=[], link_references=None, config=KACLConfig()):
        self.__data = data
        self.__headers = headers
        self.__versions = versions
        self.__link_references = link_references
        if not self.__link_references:
            self.__link_references = dict()
        self.__config = config

    def validate(self):
        """Validates the current changelog and returns KACLValidation object containing all information

        Returns:
            [KACLValidation] -- object holding all error information
        """
        validation = KACLValidation()
        # 1. assert only one header and starts on first line
        if len(self.__headers) == 0:
            validation.add_error(
                line=None,
                line_number=None,
                error_message="No 'Changelog' header found.")

            # we can stop here already
            return validation
        else:
            if self.header().raw() != self.header().raw().lstrip():
                validation.add_error(
                    line=None,
                    line_number=None,
                    error_message="Changelog header not placed on first line.")

        if len(self.__headers) > 1:
            for header in self.__headers[1:]:
                validation.add_error(
                    line=header.raw(),
                    line_number=header.line_number(),
                    error_message="Unexpected additional top-level heading found.",
                    start_character_pos=0,
                    end_character_pos=len(header.raw())
                )

        # 1.1 assert header title is in allowed list of header titles
        if self.header().title() not in self.__config.allowed_header_titles():
            header = self.header()
            start_pos = header.raw().find(header.title())
            end_pos = start_pos+len(header.title())
            validation.add_error(
                line=header.raw(),
                line_number=header.line_number(),
                error_message=f"Header title not valid. Options are [{','.join(self.__config.allowed_header_titles())}]",
                start_character_pos=start_pos,
                end_character_pos=end_pos
            )

        # 1.2 assert default content is in the header section
        for default_line in self.__config.default_content():
            if default_line not in self.header().body():
                header = self.header()
                start_pos = header.raw().find(header.title())
                end_pos = start_pos+len(header.title())
                validation.add_error(
                    line=header.raw(),
                    line_number=header.line_number(),
                    error_message=f"Missing default content '{default_line}'",
                    start_character_pos=start_pos,
                    end_character_pos=end_pos
                )

        # 2. assert 'unreleased' version is available
        # if self.get('Unreleased') == None:
        #     validation.add_error(
        #         line=None,
        #         line_number=None,
        #         error_message="'Unreleased' section is missing from the Changelog"
        #     )

        # 3. assert versions in valid format
        versions = self.versions()
        for v in versions:
            if "Unreleased" != v.version():
                raw = v.raw()
                regex = KACLParser.semver_regex
                regex_error = r'#\s+(.*)\s+'
                if v.link():
                    regex = f'#\\s+\\[{KACLParser.semver_regex}\\]'
                    regex_error = r'#\s+\[(.*)\]'
                if not KACLParser.parse_sem_ver(raw, regex):
                    start_pos = 0
                    end_pos = 0
                    m = re.match(regex_error, raw)
                    if m:
                        start_pos = raw.find(m.group(1))
                        end_pos = start_pos+len(m.group(1))
                    validation.add_error(
                        line=raw,
                        line_number=v.line_number(),
                        error_message=f"Version is not a valid semantic version.",
                        start_character_pos=start_pos,
                        end_character_pos=end_pos
                    )

        # 3.1 assert versions in descending order
        for i in range(len(versions)-1):
            try:
                v0 = versions[i]
                v1 = versions[i+1]
                if semver.compare(v0.version(), v1.version()) < 1:
                    validation.add_error(
                        line=v1.raw(),
                        line_number=v1.line_number(),
                        error_message="Versions are not in descending order.",
                        start_character_pos=0,
                        end_character_pos=len(v1.raw())
                    )
            except:
                pass

        # 3.2 assert versions have a valid date
        for v in versions:
            if "Unreleased" != v.version():
                if not v.date() or len(v.date()) < 1:
                    validation.add_error(
                        line=v.raw(),
                        line_number=v.line_number(),
                        error_message="Versions need to be decorated with a release date in the following format 'YYYY-MM-DD'",
                        start_character_pos=0,
                        end_character_pos=len(v.raw())
                    )
                if v.date() and not re.match(r'\d\d\d\d-[0-1][0-9]-[0-3][0-9]', v.date()):
                    start_pos = v.raw().find(v.date())
                    end_pos = start_pos+len(v.date())
                    validation.add_error(
                        line=v.raw(),
                        line_number=v.line_number(),
                        error_message="Date does not match format 'YYYY-MM-DD'",
                        start_character_pos=start_pos,
                        end_character_pos=end_pos
                    )

        # 3.3 check that only allowed sections are in the version
            sections = v.sections()
            for title, element in sections.items():
                if title not in self.__config.allowed_version_sections():
                    start_pos = element.raw().find(title)
                    end_pos = start_pos+len(title)
                    validation.add_error(
                        line=element.raw(),
                        line_number=element.line_number(),
                        error_message=f'"{title}" is not a valid section for a version. Options are [{",".join( self.__config.allowed_version_sections())}]',
                        start_character_pos=start_pos,
                        end_character_pos=end_pos
                    )
        # 3.4 check that only list elements are in the sections
                # 3.4.1 bring everything into a single line
                body = element.body()
                body_clean = re.sub(r'\n\s+', '', body)
                lines = body_clean.split('\n')
                non_list_lines = [x for x in lines if not x.strip(
                ).startswith('-') and len(x.strip()) > 0]
                if len(non_list_lines) > 0:
                    validation.add_error(
                        line=body.strip(),
                        line_number=element.line_number(),
                        error_message='Section does contain more than only listings.'
                    )
        # 3.5 make sure that every version that has content has it's content in a section
            if len(v.sections()) == 0 and len(v.body().strip()) != 0:
                validation.add_error(
                    line=v.raw(),
                    line_number=v.line_number(),
                    error_message=f'Version "{v.version()}" has change elements outside of a change section.'
                )

        # 3.6 Check that a link exists for linked versions
            if '[' in v.raw() and ']' in v.raw() and not v.has_link_reference():
                validation.add_error(
                    line=v.raw(),
                    line_number=v.line_number(),
                    error_message=f'Version "{v.version()}" is linked, but no link reference found in changelog file.',
                    start_character_pos=v.raw().find('['),
                    end_character_pos=v.raw().find(']')
                )

        # 4 link references
        # 4.1 check that there are only linked references
        version_strings = [v.version() for v in versions]
        for v, link in self.__link_references.items():
            if v not in version_strings:
                validation.add_error(
                    line=link.raw(),
                    line_number=link.line_number(),
                    error_message=f"Link not referenced anywhere in the document",
                    start_character_pos=0,
                    end_character_pos=len(link.raw())
                )

        return validation

    def is_valid(self):
        """Checks if the current changelog is valid
        Returns:
            [bool] -- true if valid false if not
        """
        validation_results = self.validate()
        return validation_results.is_valid()

    def has_changes(self):
        unreleased_version = self.get('Unreleased')
        if not unreleased_version:
            return False

        sections = unreleased_version.sections()
        if not sections or len(sections) == 0:
            return False

        for pair in sections.items():
            if pair[1] and len(pair[1].items()) > 0:
                return True

        return False

    def add(self, section, data):
        """adds a new change to a given section in the 'unreleased' version

        Arguments:
            section {[str]} -- section to add data to
            data {[str]} -- change information
        """
        unreleased_version = self.get('Unreleased')
        if unreleased_version == None:
            unreleased_version = KACLVersion(version="Unreleased")
            self.__versions.insert(0, unreleased_version)
        unreleased_version.add(section.capitalize(), data)

    def release(self, version=None, link=None, auto_link=False, increment=None):
        """Creates a new release version by copying the 'unreleased' changes into the
        new version

        Keyword Arguments:
            link {[str]} -- url the version will be linked with (default: {None})
            version {[str]} -- semantic versioning string
            increment {[str]} -- use either 'patch', 'minor', or 'major' to automatically increment the last version
        """
        if increment:
            v = self.current_version()
            if v:
                sv = semver.VersionInfo.parse(v)
                if 'patch' == increment:
                    sv = sv.bump_patch()
                elif 'minor' == increment:
                    sv = sv.bump_minor()
                elif 'major' == increment:
                    sv = sv.bump_major()
                version = str(sv)
            else:
                raise KACLException("No previously released version found. Incrementing not possible")

        # check that version is a valid semantic version
        semver.parse(version) # --> will throw a ValueError if version is not a valid semver

        # check if there are changes to release
        if self.has_changes() is False:
            raise KACLException("The current changelog has no changes. You can only release if changes are available.")

        # check if the version already exists
        if self.get(version) != None:
            raise KACLException(f"The version '{version}' already exists in the changelog. You cannot release the same version twice.")

        # check if new version is greater than the last one
        #   1. there has to be an 'unreleased' section
        #   2. All other versions are in descending order
        version_list = self.versions()
        if len(version_list) > 1: # versions[0] --> unreleased
            last_version = version_list[1].version()
            if semver.compare(version, last_version) < 1:
                raise KACLException(f"The version '{version}' cannot be released since it is smaller than the preceding version '{last_version}'.")

        # get current unreleased changes
        unreleased_version = self.get('Unreleased')

        # remove current unrelease version from list
        self.__versions.pop(0)

        self.__versions.insert(0, KACLVersion(version="Unreleased", link=unreleased_version.link()))

        # convert unreleased version to version
        self.__versions.insert(1, KACLVersion(version=version,
                                              link=KACLElement(
                                                  title=version, body=link),
                                              date=datetime.datetime.now().strftime("%Y-%m-%d"),
                                              sections=unreleased_version.sections()))

        if auto_link:
            link_provider = self.__get_link_provider()
            for i in range(2):
                fargs = {
                    "version": self.__versions[i].version(),
                    "previous_version": None,
                    "latest_version": version
                }

                if len(self.__versions) > i+1:
                    fargs["previous_version"] = self.__versions[i+1].version()

                if 'unreleased' in self.__versions[i].version().lower():
                    self.__versions[i].set_link( link_provider.unreleased_changes(**fargs) )
                else:
                    if fargs["previous_version"]:
                        self.__versions[i].set_link( link_provider.compare_versions(**fargs) )
                    else:
                        self.__versions[i].set_link( link_provider.initial_version(**fargs) )

    def get(self, version):
        """Returns the selected version

        Arguments:
            version {[str]} -- semantic versioning string

        Returns:
            [KACLVersion] -- version object with all information
        """
        res = [x for x in self.__versions if x.version()
               and version.capitalize() == x.version()]
        if res and len(res):
            return res[0]

    def current_version(self):
        """returns the current version (last released)

        Returns:
            [str] -- latest released version, None if none is available
        """
        version_list = self.versions()
        for v in version_list:
            if v.version().lower() != 'unreleased':
                return v.version()

    def generate_links(self, host_url=None, compare_versions_template=None, unreleased_changes_template=None, initial_version_template=None):
        """automatically generates links for all versions

        Returns: None
        """
        link_provider = self.__get_link_provider(host_url=host_url,
                                                 compare_versions_template=compare_versions_template,
                                                 unreleased_changes_template=unreleased_changes_template,
                                                 initial_version_template=initial_version_template)

        versions = self.versions()
        if len(versions) > 1:
            for i in range(len(versions)-1):
                fargs = {
                    "version": versions[i].version(),
                    "previous_version": versions[i+1].version(),
                    "latest_version": self.current_version()
                }

                if 'unreleased' in versions[i].version().lower():
                    versions[i].set_link( link_provider.unreleased_changes(**fargs) )
                else:
                    versions[i].set_link( link_provider.compare_versions(**fargs) )
            versions[-1].set_link( link_provider.initial_version(**fargs) )
        elif len(versions) == 1:
            fargs = {
                "version": versions[0].version(),
                "latest_version": self.current_version()
            }

            if 'unreleased' in versions[0].version().lower():
                versions[0].set_link( link_provider.initial_version(version="master") )
            else:
                versions[0].set_link( link_provider.initial_version(**fargs) )


    def header(self):
        """Gives access to the top level heading element

        Returns:
            [KACLElement] -- object holding all information of the top level heading
        """
        if self.__headers and len(self.__headers) > 0:
            return self.__headers[0]

    def title(self):
        """Returns the title of the changelog

        Returns:
            [str] -- title of the changelog
        """
        if self.__headers and len(self.__headers) > 0:
            return self.__headers[0].title()
        return None

    def versions(self):
        """Returns a list of all available versions

        Returns:
            [list] -- list of KACLVersions
        """
        return self.__versions

    def set_config(self, config):
        self.__config = config

    def config(self):
        return self.__config

    @staticmethod
    def init():
        return KACLDocument.parse("""# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/), and this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).

## Unreleased
        """)

    @staticmethod
    def parse(data):
        """Parses a given text object and returns the KACLDocument

        Arguments:
            data {[str]} -- markdown text holding the changelog

        Returns:
            [KACLDocument] -- object holding all information
        """

        data_lf = data.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)

        # First check if there are link references and split the document where they begin
        link_reference_begin, link_references = KACLParser.parse_link_references(data_lf)

        changelog_body = data_lf
        if link_reference_begin:
            changelog_body = data_lf[:link_reference_begin]

        # read header
        headers = KACLParser.parse_header(changelog_body, 1, 2)

        # read versions
        versions = KACLParser.parse_header(changelog_body, 2, 2)
        versions = [KACLVersion(element=x) for x in versions]

        # set link references into versions if available
        for v in versions:
            v.set_link(link_references.get(v.version(), None))

        return KACLDocument(data=data, headers=headers, versions=versions, link_references=link_references)


    def __get_link_provider(self, host_url=None, compare_versions_template=None, unreleased_changes_template=None, initial_version_template=None):
        host_url = host_url if host_url else self.__config.link_host_url()
        compare_versions_template = compare_versions_template if compare_versions_template else self.__config.links_compare_versions_template()
        unreleased_changes_template = unreleased_changes_template if unreleased_changes_template else self.__config.links_unreleased_changes_template()
        initial_version_template = initial_version_template if initial_version_template else self.__config.links_initial_version_template()

        if host_url is None:
            if 'CI_PROJECT_URL' in os.environ:
                host_url = os.environ['CI_PROJECT_URL']
            else:
                try:
                    repo = git.Repo(os.getcwd())
                    remote = repo.remote()
                    for url in remote.urls:
                        host_url = url
                        break
                except:
                    raise KACLException("ERROR: Could not determine project url. Update your config or run within a valid git repository")

        return LinkProvider(host_url=host_url,
                            compare_versions_template=compare_versions_template,
                            unreleased_changes_template=unreleased_changes_template,
                            initial_version_template=initial_version_template)



from setuptools import setup
from os import path
import re

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

version = re.search(r'^__version__\s*=\s*"(.*)"',
                    open('kacl/__init__.py').read(),
                    re.M).group(1)

setup(name='python-kacl',
      version=version,
      description='Python module and CLI tool for validating and modifying Changelogs in "keep-a-changelog" format"',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/mschmieder/python-kacl',
      author='Matthias Schmieder',
      author_email='schmieder.matthias@gmail.com',
      entry_points={
           "console_scripts": ['kacl-cli = kacl.kacl_cli:start']
      },
      license='MIT',
      packages=['kacl'],
      include_package_data=True,
      python_requires='>=3.6',
      install_requires=[
          'click',
          'semver',
          'pychalk',
          'gitpython',
          'pyyaml',
          'python-box'
      ],
      zip_safe=False,
      classifiers= [
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.6",
          "Intended Audience :: Developers",
          "Topic :: Software Development :: Version Control"
      ])



# Version of the python-kacl package
__version__ = "0.2.26"

from .document import *
from .serializer import *

def load(file):
    """
    Parse the first YAML document in a stream
    and produce the corresponding Python object.
    """
    doc = None
    with open(file, 'r') as f:
        document = f.read()
        try:
            doc = KACLDocument.parse(document)
        finally:
            f.close()
    return doc

def parse(text):
    return KACLDocument.parse(text)

def dump(document):
    return KACLMarkdownSerializer().serialize(document)

def new():
    return KACLDocument.init()